#include <GL/glew.h>
#include <GL/freeglut.h>
#include <GL/glut.h>
#include <stdio.h>
#include <string.h>

// Súradnice vertexov a farby
static const GLfloat vertexBufferData[] = {
	-1, -1,
	-1,  1,
	 1, -1,
	 1,  1,
};
GLuint vertexBuffer;

// Kód vertex a fragment shadera
static const char vertexShader[] = "#version 330 core\n\
\n\
layout(location = 0) in vec2 position;\n\
\n\
void main(){\n\
	gl_Position = vec4(position, 0.0, 1.0);\n\
}\n\
";
char *fragmentShader;
GLuint shaderProgram;

static const char fragmentShaderHeader[] = "#version 330 core\n\
\n\
const vec2 iResolution = vec2(800.0, 480.0);\n\
\n\
out vec4 outColor;\n\
\n\
void mainImage(out vec4 fragColor, in vec2 fragCoord);\n\
\n\
void main()\n\
{\n\
	mainImage(outColor, gl_FragCoord.xy);\n\
}\n";

// Výpis chyby kompilácie ak shader nebol skompilovaný
void printShaderStatus(GLuint shader) {
	GLint res = GL_FALSE;
	int infoLength;

	glGetShaderiv(shader, GL_COMPILE_STATUS, &res);
	glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infoLength);
	if (res != GL_TRUE) {
		char *msg = malloc(infoLength + 1);
		glGetShaderInfoLog(shader, infoLength, NULL, msg);
		printf("%s", msg);
		free(msg);
	}
}

// Výpis chyby linkovania ak sa program nepodarilo zlinkovať
void printProgramStatus(GLuint program) {
	GLint res = GL_FALSE;
	int infoLength;

	glGetProgramiv(program, GL_LINK_STATUS, &res);
	glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infoLength);
	if (res != GL_TRUE) {
		char *msg = malloc(infoLength + 1);
		glGetProgramInfoLog(program, infoLength, NULL, msg);
		printf("%s", msg);
		free(msg);
	}
}

// Kompilácia ľubovoľného druhu shadera
void compileShader(GLuint shader, const char *code) {
	glShaderSource(shader, 1, &code, NULL);
	glCompileShader(shader);
	printShaderStatus(shader);
}

// Kompilácia shaderov a zlinkovanie do programu
GLuint getShaderProgram(const char *vertexShaderSource, const char *fragmentShaderSource) {
	// Vytvorenie programu
	GLuint program = glCreateProgram();
	// Vytvorenie inštancií shaderov
	GLuint vertexShader = glCreateShader(GL_VERTEX_SHADER);
	GLuint fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
	// Kompilácia oboch zdrojových kódov
	compileShader(vertexShader, vertexShaderSource);
	compileShader(fragmentShader, fragmentShaderSource);
	// Priradenie shaderov k jednému programu
	glAttachShader(program, vertexShader);
	glAttachShader(program, fragmentShader);
	// Zlinkovanie programu a výpis chyb ak nebolo linkovanie úspešné
	glLinkProgram(program);
	printProgramStatus(program);
	// Ďalej shader nnemusí byť priradený k programu
	glDetachShader(program, vertexShader);
	glDetachShader(program, fragmentShader);
	return program;
}


// Zobrazenie štvorca
void display() {
	// Vyčistenie pozadia
	glClear(GL_COLOR_BUFFER_BIT);

	// Pri vykresľovaní sa použije shader
	glUseProgram(shaderProgram);

	// Priradenie poľa k atribútu s location(0) (position)
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
	glVertexAttribPointer(
		0, // Index atribútu (zodpovedá hodnote v location)
		2, // Počet komponentov (zadávame pre každý bod iba 2 súradnice)
		GL_FLOAT, // V poli sú hodnoty s pohyblivou desatinnou čiarkou
		GL_FALSE, // Nemá význam pre typ float
		0, // Rozostup medzi prvkami
		(void *)0 // Posun voči začiatku poľa
	);

	// Vykreslenie objektu triangle strip zo 4 vertexov
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);

	// Zobrazenie na obrazovke
	glFlush();
}

// Inicializácia OpenGL a hlavnej slučky
int main(int argc, char** argv) {
	if (argc != 2) {
		fprintf(stderr, "Path with fragment shader required\n");
		return -1;
	}

	FILE *fp = fopen(argv[1], "rb");
	if (!fp) {
		fprintf(stderr, "Can't read file: %s", argv[1]);
		return -1;
	}
	fseek(fp, 0, SEEK_END);
	size_t shaderSize = ftell(fp);
	size_t headerSize = strlen(fragmentShaderHeader);
	fseek(fp, 0, SEEK_SET);

	// Ak je kompletný shader, nepridáva sa hlavička
	char version[8];
	fread(version, 8, 1, fp);
	if (strncmp(version, "#version", 8) == 0) {
		headerSize = 0;
	}

	fragmentShader = malloc(shaderSize + headerSize + 1);
	if (headerSize) { // Kopírovanie štandardnej hlavičky
		strcpy(fragmentShader, fragmentShaderHeader);
	}

	fseek(fp, 0, SEEK_SET);
	fread(fragmentShader + headerSize, shaderSize, 1, fp);
	fclose(fp);
	fragmentShader[shaderSize + headerSize] = '\0';

	glutInit(&argc, argv);

	// Explicitne vyžadujem OpenGL 3.3 core profile
	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);

	// Režim RGBA s double bufferom
	glutInitDisplayMode(GLUT_SINGLE | GLUT_RGBA);

	// Vytvorenie okna
	glutInitWindowSize(800, 480);
	glutCreateWindow(argv[1]);

	// Inicializácia GLEW musí nasledovať po otvorení okna
	glewInit();

	// Aktivovaný vertex array objekt
	GLuint vertexArray;
	glGenVertexArrays(1, &vertexArray);
	glBindVertexArray(vertexArray);

	// Kompilácia
	shaderProgram = getShaderProgram(vertexShader, fragmentShader);

	// Upload vertexov a farieb do GPU
	glGenBuffers(1, &vertexBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertexBufferData), vertexBufferData, GL_STATIC_DRAW);

	// Zaregistrovanie funkcie, ktorá vykreslí obsah okna
	glutDisplayFunc(display);

	// Hlavná slučka
	glutMainLoop();

	free(fragmentShader);
}
